/*

    File: rnntest.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
//DJGPP is our djgpp gnu c++ compiler flag
//WIN32 is our MSVC flag?

#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <assert.h>		//i wish djgpp assertion failures would do a stack dump

#ifdef DJGPP	//djgpp-only goods
#include <pc.h>         //inport
#endif

#ifdef WIN32	//do windows code for all the console calls

#include <windows.h>
void gotoxy(int x, int y) {
	//DOS starts at 1, windows COORDs start at 0
	COORD pos = {x-1,y-1};
	SetConsoleCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE), pos);
}

void clrscr(void) {
	HANDLE hConsoleOut = GetStdHandle(STD_OUTPUT_HANDLE);

	//while we're at it...
	COORD newSize = {80, 50};
	SetConsoleScreenBufferSize(hConsoleOut, newSize);

//this clears the screen ... but its already cleared in the beginning, the only time this method is called...
//so whats the point?
//
//	CONSOLE_SCREEN_BUFFER_INFO screenBufferInfo;
//	GetConsoleScreenBufferInfo(hConsoleOut, &screenBufferInfo);
//	DWORD size = screenBufferInfo.dwSize.X * screenBufferInfo.dwSize.Y;
//	COORD pos = {0,0};
//	DWORD truesize;
//	FillConsoleOutputCharacter(hConsoleOut, ' ', size, pos, &truesize);
//	gotoxy(1,1);
}

#define numberof(x)			(sizeof(x)/sizeof(*(x)))

char getkeystate(void) {
	INPUT_RECORD inputRecord[1];

	HANDLE hConsoleIn = GetStdHandle(STD_INPUT_HANDLE);
	int size = numberof(inputRecord);

	DWORD truesize;
	PeekConsoleInput(hConsoleIn, inputRecord, size, &truesize);
	if (!truesize) return 0;

	INPUT_RECORD *in = &inputRecord[0];
	if (in->EventType != KEY_EVENT) return 0;

	KEY_EVENT_RECORD *key = &in->Event.KeyEvent;

	char sc = (char)key->wVirtualScanCode;
	if (!key->bKeyDown) sc = !sc;
	return sc;
}

#endif

#ifdef DJGPP
char getkeystate(void) {	return inportb(0x60); }
#endif

//#define FORCE_STEPS		2	//1, 2, or undef are valid values - how many layers to force the net to be
#define	ALLOW_HISTORY_WEIGHT_INFLUENCE

#define RandomNum()     (((double)rand() / (double)RAND_MAX) * 2.0 - 1.0)
#define mabs(x) (x<0?-x:x)

static size_t total_allocated = 0;
void *operator_new(size_t size) {
	total_allocated += size;
	return malloc(size);
	//HOW CAN I OVERLOADTHIS FOR ALL OPERATOR NEW'S?! IN DJGPP?
}

double *new_double(int count) {
	total_allocated += sizeof(double) * count;
	return new double[count];
}

double **new_double_ptr(int count) {
	total_allocated += sizeof(double *) * count;
	return new double*[count];
}

double ***new_double_ptr_ptr(int count) {
	total_allocated += sizeof(double **) * count;
	return new double**[count];
}

template<class T>
class CVector {
public:
	int size;
	T *v;

	void *operator new(size_t size) { return operator_new(size);}

	CVector(void) {
		size = 0;
		v = NULL;
	}

	CVector(int _size) {
		size = _size;
		v = new T[size];
		assert(v);
		memset(v, 0, sizeof(T) * size);
	}

	~CVector() {
		if (v) delete v;
	}

	void Resize(int newsize) {
		int i;

		T *v2 = new T[newsize];
		assert(v2);
		
		int min_size = newsize;
		if (min_size > size) min_size = size;
		for (i = 0; i < min_size; i++) {
			v2[i] = v[i];
		}
		for (; i < newsize; i++) {
			v2[i] = 0;
		}

		if (v) delete v;

		v = v2;
		size = newsize;
	}

	void Fill(const CVector &src) {
		int min_size = src.size;
		if (min_size > size) min_size = size;

		for (int i = 0; i < min_size; i++) {
			v[i] = src.v[i];
		}
	}

	void Fill(T val) {
		for (int i = 0; i < size; i++) {
			v[i] = val;
		}
	}

	T &operator[](int index) {
		assert (index >= 0 && index < size);
		return v[index];
	}

	T operator[](int index) const {
		assert (index >= 0 && index < size);
		return v[index];
	}
};

typedef CVector<double> CVectord;

class CMatrix {
public:
	int size1, size2;
	double **v;

	void *operator new(size_t size) { return operator_new(size);}

	CMatrix(void) {
		size1 = 0;
		size2 = 0;
		v = NULL;
	}

	CMatrix(int _size1, int _size2) {
		size1 = _size1;
		size2 = _size2;
		v = new_double_ptr(size1);
		assert(v);
		memset(v, 0, sizeof(double*) * size1);
		for (int i = 0; i < size1; i++) {
			v[i] = new_double(size2);
			assert(v[i]);
			memset(v[i], 0, sizeof(double) * size2);
		}
	}

	~CMatrix() {
		if (v) {
			for (int i = 0; i < size1; i++) {
				if (v[i]) delete v[i];
			}
			delete v;
		}
	}

	void Fill(const CMatrix &src) {
		int min_size1 = src.size1;
		if (min_size1 > size1) min_size1 = size1;
		int min_size2 = src.size2;
		if (min_size2 > size2) min_size2 = size2;

		for (int j = 0; j < min_size2; j++) {
			for (int i = 0; i < min_size1; i++) {
				this->v[i][j] = src.v[i][j];
			}
		}
	}

	void Fill(double val) {
		for (int j = 0; j < size2; j++) {
			for (int i = 0; i < size1; i++) {
				this->v[i][j] = val;
			}
		}
	}

	void Resize(int newsize1, int newsize2) {
		int i,j;

		double **v2 = new_double_ptr(newsize1);
		if (!v2) return;
		memset(v2, 0, sizeof(double*) * newsize1);
		for (i = 0; i < newsize1; i++) {
			v2[i] = new_double(newsize2);
			assert(v2[i]);
			memset(v2[i], 0, sizeof(double) * newsize2);
		}

//		int min_size1 = newsize1;
//		if (min_size1 > size1) min_size1 = size1;
//		int min_size2 = newsize2;
//		if (min_size2 > size2) min_size2 = size2;

		for (j = 0; j < newsize2; j++) {
			for (i = 0; i < newsize1; i++) {
				if (i < size1 && j < size2) {
					v2[i][j] = v[i][j];
				}
			}
		}

		if (v) {
			for (i = 0; i < size1; i++) {
				if (v[i]) delete v[i];
			}
			delete v;
		}

		v = v2;
		size1 = newsize1;
		size2 = newsize2;
	}

	class safe_accessor {
	public:
		double *vi;
		int size2;
		safe_accessor(double *_vi, int _size2) {
			vi = _vi;
			size2 = _size2;
		}
		double &operator[](int j) {
			assert (j >= 0 && j < size2);
			return vi[j];
		}
	};

	safe_accessor operator[](int i) {
		assert (i >= 0 && i < size1);
		return safe_accessor(v[i], size2);
	}

};

//rank 3 tensor - used for storing net history influence coefficients
class CTensor3 {
public:
	int size1, size2, size3;
	double ***v;

	void *operator new(size_t size) { return operator_new(size);}

	CTensor3(void) {
		size1 = 0;
		size2 = 0;
		size3 = 0;
		v = NULL;
	}

	CTensor3(int _size1, int _size2, int _size3) {
		size1 = _size1;
		size2 = _size2;
		size3 = _size3;
		v = new_double_ptr_ptr(size1);
		assert(v);
		memset(v, 0, sizeof(double**) * size1);
		for (int i = 0; i < size1; i++) {
			v[i] = new_double_ptr(size2);
			assert(v[i]);
			memset(v[i], 0, sizeof(double*) * size2);
			for (int j = 0; j < size2; j++) {
				v[i][j] = new_double(size3);
				assert(v[i][j]);
				memset(v[i][j], 0, sizeof(double) * size3);
			}
		}
	}

	~CTensor3() {
		if (v) {
			for (int i = 0; i < size1; i++) {
				if (v[i]) {
					for (int j = 0; j < size2; j++) {
						if (v[i][j]) delete v[i][j];
					}
					delete v[i];
				}
			}
			delete v;
		}
	}

	void Fill(const CTensor3 &src) {
		int min_size1 = src.size1;
		if (min_size1 > size1) min_size1 = size1;
		int min_size2 = src.size2;
		if (min_size2 > size2) min_size2 = size2;
		int min_size3 = src.size3;
		if (min_size3 > size3) min_size3 = size3;

		for (int k = 0; k < min_size3; k++) {
			for (int j = 0; j < min_size2; j++) {
				for (int i = 0; i < min_size1; i++) {
					this->v[i][j][k] = src.v[i][j][k];
				}
			}
		}
	}

	void Resize(int newsize1, int newsize2, int newsize3) {
		int i,j,k;

		double ***v2 = new_double_ptr_ptr(newsize1);
		assert(v2);
		memset(v2, 0, sizeof(double**) * newsize1);
		for (i = 0; i < newsize1; i++) {
			v2[i] = new_double_ptr(newsize2);
			assert(v2[i]);
			memset(v2[i], 0, sizeof(double*) * newsize2);
			for (j = 0; j < newsize2; j++) {
				v2[i][j] = new_double(newsize3);
				assert(v2[i][j]);
				memset(v2[i][j], 0, sizeof(double) * newsize3);
			}
		}

		for (k = 0; k < newsize3; k++) {
			for (j = 0; j < newsize2; j++) {
				for (i = 0; i < newsize1; i++) {
					if (i < size1 && j < size2 && k < size3) {
						v2[i][j][k] = v[i][j][k];
					}
				}
			}
		}

		if (v) {
			for (i = 0; i < size1; i++) {
				if (v[i]) {
					for (j = 0; j < size2; j++) {
						if (v[i][j]) delete v[i][j];
					}
					delete v[i];
				}
			}
			delete v;
		}

		v = v2;
		size1 = newsize1;
		size2 = newsize2;
		size3 = newsize3;
	}

	//might as well use multiple vectors per dimension...

	class safe_accessor2 {
	public:
		double *vj;
		int size3;
		safe_accessor2(double *_vj, int _size3) {
			vj = _vj;
			size3 = _size3;
		}
		double &operator[](int k) {
			assert (k >= 0 && k < size3);
			return vj[k];
		}
	};

	class safe_accessor {
	public:
		double **vi;
		int size2;
		int size3;
		safe_accessor(double **_vi, int _size2, int _size3) {
			vi = _vi;
			size2 = _size2;
			size3 = _size3;
		}
		safe_accessor2 operator[](int j) {
			assert (j >= 0 && j < size2);
			return safe_accessor2(vi[j], size3);
		}
	};

	safe_accessor operator[](int i) {
		assert (i >= 0 && i < size1);
		return safe_accessor(v[i], size2, size3);
	}
};

double sigmoid(double x) {
	const double coeff_const = 5.0;
	return 1.0/(1.0 + exp(-coeff_const * x));
}

class C_RNNLayer {
public:

	void *operator new(size_t size) { return operator_new(size);}

	//the neuron signals
	CVectord in;
	CVectord mem;
	CVectord out;
	CVectord des;

	//the connecting weights
	CVectord mem_bias;
	CMatrix in_mem;
	CMatrix mem_mem;

	CVectord out_bias;
	CMatrix in_out;
	CMatrix mem_out;

	//the history information
	CMatrix h_mem_bias;
	CTensor3 h_in_mem;
	CTensor3 h_mem_mem;

	C_RNNLayer(int in_size, int mem_size, int out_size) {
		in.Resize(in_size);
		mem.Resize(mem_size);
		out.Resize(out_size);
		des.Resize(out_size);

		mem_bias.Resize(mem_size);
		in_mem.Resize(in_size, mem_size);
		mem_mem.Resize(mem_size, mem_size);

		out_bias.Resize(out_size);
		in_out.Resize(in_size, out_size);
		mem_out.Resize(mem_size, out_size);

		h_mem_bias.Resize(mem_size,mem_size);			//[target mem][iterated mem]
		h_in_mem.Resize(in_size,mem_size,mem_size);		//[target in][target mem][iterated mem]
		h_mem_mem.Resize(mem_size,mem_size,mem_size);	//[target mem][target mem][iterated mem]
	}

	void RandomizeWeights(double range) {
		int i,j;
		for (j = 0; j < mem.size; j++) {
			mem_bias[j] = RandomNum() * range;

			for (i = 0; i < in.size; i++) {
				in_mem[i][j] = RandomNum() * range;
			}

			for (i = 0; i < mem.size; i++) {
				mem_mem[i][j] = RandomNum() * range;
			}
		}

		for (j = 0; j < out.size; j++) {
			out_bias[j] = RandomNum() * range;

			for (i = 0; i < in.size; i++) {
				in_out[i][j] = RandomNum() * range;
			}

			for (i = 0; i < mem.size; i++) {
				mem_out[i][j] = RandomNum() * range;
			}
		}
	}

	//feeds the network forward
	void Evaluate(CVectord &new_mem) {

		int i,j;

		//if we are told to evaluate a new state...
		if (&new_mem) {

			//grab the smallest size between the two 
			int size = new_mem.size;
			if (mem.size < new_mem.size) size = mem.size;

			for (j = 0; j < size; j++) {
				new_mem[j] = mem_bias[j];
	
				for (i = 0; i < in.size; i++) {
					new_mem[j] += in_mem[i][j] * in[i];
				}
	
				for (i = 0; i < mem.size; i++) {
					new_mem[j] += mem_mem[i][j] * mem[i];
				}
	
				new_mem[j] = sigmoid(new_mem[j]);
			}
		}

		for (j = 0; j < out.size; j++) {
			out[j] = out_bias[j];

			for (i = 0; i < in.size; i++) {
				out[j] += in_out[i][j] * in[i];
			}

			for (i = 0; i < mem.size; i++) {
				out[j] += mem_out[i][j] * mem[i];
			}

			out[j] = sigmoid(out[j]);
		}
	}

	void FillWeights(C_RNNLayer *src) {
		out_bias.Fill	(src->out_bias);
		in_out.Fill		(src->in_out);
		mem_out.Fill	(src->mem_out);
		mem_bias.Fill	(src->mem_bias);
		in_mem.Fill		(src->in_mem);
		mem_mem.Fill	(src->mem_mem);
	}

	void PrintWeight(double v) {
		printf("%c", v < 0 ? '-' : '+');
	}

	void PrintSignal(double v) {
//		printf("%c", v < 0.5 ? '0' : '1');
		if (v < 0.2) {
			printf(" ");
		} else if (v < 0.4) {
			printf("%c", 176);
		} else if (v < 0.6) {
			printf("%c", 177);
		} else if (v < 0.8) {
			printf("%c", 178);
		} else {
			printf("%c", 219);
		}
	}

	void Display(CVectord *new_mem = NULL) {
		int a,b;

		gotoxy(1,1);

//		printf("       mem_bias | out_bias\n");
//		printf("in  -|  in_mem  |  in_out\n");
//		printf("mem -| mem_mem  | mem_out\n");
//		printf("     +--------------------\n");
//		printf("           mem  |     out\n");
//		printf("\n");

		printf(" |");
		for (a = 0; a < mem_bias.size; a++) {
			PrintWeight(mem_bias[a]);
		}
		printf("|");
		for (a = 0; a < out_bias.size; a++) {
			PrintWeight(out_bias[a]);
		}
		printf("|\n");
		printf("\n");

		for (b = 0; b < in.size; b++) {
			PrintSignal(in[b]);
			printf ("|");
			for (a = 0; a < in_mem.size2; a++) {
				PrintWeight(in_mem[b][a]);
			}
			printf("|");
			for (a = 0; a < in_out.size2; a++) {
				PrintWeight(in_out[b][a]);
			}
			printf("|\n");
		}
		printf("\n");

		for (b = 0; b < mem.size; b++) {
			PrintSignal(mem[b]);
			printf ("|");
			for (a = 0; a < mem_mem.size2; a++) {
				PrintWeight(mem_mem[b][a]);
			}
			printf("|");
			for (a = 0; a < mem_out.size2; a++) {
				PrintWeight(mem_out[b][a]);
			}
			printf("|\n");
		}
		printf("\n");

		printf(" |");
		for (a = 0; a < mem.size; a++) {
			if (new_mem) {
				PrintSignal(new_mem->v[a]);
			} else {
				printf(" ");
			}
		}
		printf("|");
		for (a = 0; a < out.size; a++) {
			PrintSignal(out[a]);
		}
		printf("|\n");
		printf("\n");

		printf(" |");
		for (a = 0; a < mem.size; a++) {
			printf(" ");
		}
		printf("|");
		for (a = 0; a < des.size; a++) {
			PrintSignal(des[a]);
		}
		printf("|\n");
		printf("\n");
	}

	void PrintFloat(double v, int prec = 2) {
		if (v >= 0) {
			printf(" ");
		} else {
			printf("-");
			v = -v;
		}
		char printStr[256];
		sprintf(printStr, "%%1.%df", prec);
		printf(printStr, v);
	}
	void PrintSpace(void) {
		printf("    ");
	}

	void DisplayFull(CVectord *new_mem = NULL) {
		int a,b;

		gotoxy(1,1);

//		printf("       mem_bias | out_bias\n");
//		printf("in  -|  in_mem  |  in_out\n");
//		printf("mem -| mem_mem  | mem_out\n");
//		printf("     +--------------------\n");
//		printf("           mem  |     out\n");
//		printf("\n");

		PrintSpace();
		printf("|");
		for (a = 0; a < mem_bias.size; a++) {
			PrintFloat(mem_bias[a]);
		}
		printf("|");
		for (a = 0; a < out_bias.size; a++) {
			PrintFloat(out_bias[a]);
		}
		printf("|\n");
		printf("\n");

		for (b = 0; b < in.size; b++) {
			PrintFloat(in[b]);
			printf ("|");
			for (a = 0; a < in_mem.size2; a++) {
				PrintFloat(in_mem[b][a]);
			}
			printf("|");
			for (a = 0; a < in_out.size2; a++) {
				PrintFloat(in_out[b][a]);
			}
			printf("|\n");
		}
		printf("\n");

		for (b = 0; b < mem.size; b++) {
			PrintFloat(mem[b]);
			printf ("|");
			for (a = 0; a < mem_mem.size2; a++) {
				PrintFloat(mem_mem[b][a]);
			}
			printf("|");
			for (a = 0; a < mem_out.size2; a++) {
				PrintFloat(mem_out[b][a]);
			}
			printf("|\n");
		}
		printf("\n");

		PrintSpace();
		printf("|");
		for (a = 0; a < mem.size; a++) {
			if (new_mem) {
				PrintFloat(new_mem->v[a]);
			} else {
				PrintSpace();
			}
		}
		printf("|");
		for (a = 0; a < out.size; a++) {
			PrintFloat(out[a]);
		}
		printf("|\n");
		printf("\n");

		PrintSpace();
		printf("|");
		for (a = 0; a < mem.size; a++) {
			PrintSpace();
		}
		printf("|");
		for (a = 0; a < des.size; a++) {
			PrintFloat(des[a]);
		}
		printf("|\n");
		printf("\n");
	}
};

typedef CVector<C_RNNLayer *> CVectorNet;

class C_RNN {
public:

	void *operator new(size_t size) { return operator_new(size);}

	//vector of networks
	CVectorNet n;

	//the current network (in the circular buffer)
	int index;

	C_RNN(int history, int in_size, int mem_size, int out_size) {

		//make sure we at least have nets for:
		//the current state, the next state,
		//and as many prev states needed for the gradient
		if (history < 2) history = 2;

		//give us an extra out neuron for "ready with results"
		out_size++;

		//and an extra in neuron for "got input"
		in_size++;

		//create the net vector
		n.Resize(history);

		//and fill it
		for (int i = 0; i < history; i++) {

			//plus one out_size for the ready signal
			n[i] = new C_RNNLayer(in_size, mem_size, out_size);
		}

		//set the current network to zero.
		index = 0;

		//fill the initial weights with random values
		n[0]->RandomizeWeights(0.1);

		//leave the initial state at zero...?

	}

	~C_RNN() {
		for (int i = 0; i < n.size; i++) {
			if (n[i]) delete n[i];
		}
	}

	//adjusts weights based upon the network of the current index
	//depends upon ComputeHistoryTensor's contents
	//also depends on next_net to have a copy fo the last net's weights
	void AdjustWeights(void) {

		int i,j,j2,j3,k;
		double sum;

		int next_index = (index + 1) % n.size;
		C_RNNLayer *next_net = n[next_index];

		//based upon error, and history vector (all but current & next network)
		// - calculate new weights & store in next network
		//perform gradient descent on each weight
		//where our error is sum of difference between desired and actual
		//outputs, squared (and halved for coefficient's sake) over all time

		//decay a coefficient here pertaining to the influence
		//so that newer generations reflect greater changes
		//NOTICE - if infl_decay is too high then the SOLUTION WILL DIVERGE!
		const double infl_decay = 0.25;
		//the initial influence can also be seen as the weight change coefficient
		double influence = 1.0;

		for (int t = index; t != next_index; t = ((t + n.size-1) % n.size)) {
			C_RNNLayer *nt = n[t];

			//first the output gradients - the easy ones...

			for (k = 0; k < next_net->out.size; k++) {

				//// out_bias

				//delta = (d - y)y(1 - y)
				//where y(1-y) is the sigmoid derivitive

				//solve dE/dwbyk (bias weight for the k'th output)
				double delta =
					 (nt->des[k] - nt->out[k])
					 * nt->out[k]
					 * (1.0 - nt->out[k]);

				//continually offset
				next_net->out_bias[k] += influence * delta;

				//// in_out

				for (i = 0; i < next_net->in.size; i++) {
					next_net->in_out[i][k] += influence * delta * nt->in[i];
				}

				//// mem_out

				for (j = 0; j < next_net->mem.size; j++) {
					next_net->mem_out[j][k] += influence * delta * nt->mem[j];
				}
			}


			//next the memory gradients - the recursive ones...

			for (k = 0; k < next_net->out.size; k++) {
				double delta = 
					 (nt->des[k] - nt->out[k])
					 * nt->out[k]
					 * (1.0 - nt->out[k]);

				for (j = 0; j < next_net->mem.size; j++) {
					double delta2 = nt->mem_out[j][k]
						* nt->mem[j]
						* (1.0 - nt->mem[j]);

					//// mem_bias

					//accumulate derivative components from the memory layer recursion
					sum = 0;
					for (j2 = 0; j2 < next_net->mem.size; j2++) {
						sum += nt->mem_out[j2][k] * nt->h_mem_bias[j][j2];
					}
					next_net->mem_bias[j] += influence * delta * sum;

					//// in_mem

					for (i = 0; i < next_net->in.size; i++) {

						//accumulate derivative components from the memory layer recursion
						sum = 0;
						for (j2 = 0; j2 < next_net->mem.size; j2++) {
							sum += nt->mem_out[j2][k] * nt->h_in_mem[i][j][j2];
						}
						next_net->in_mem[i][j] += influence * delta * sum;
					}

					//// mem_mem

					for (j2 = 0; j2 < next_net->mem.size; j2++) {
						//accumulate derivative components from the memory layer recursion
						sum = 0;
						for (j3 = 0; j3 < next_net->mem.size; j3++) {
							sum += nt->mem_out[j3][k] * nt->h_mem_mem[j][j2][j3];
						}
						//i swapped j and j2 - maybe i shouldnt have?  my math says [j][j2] ...
						next_net->mem_mem[j][j2] += influence * delta * sum;
					}
				}
			}


			//decay over time
			influence *= infl_decay;

#ifndef ALLOW_HISTORY_WEIGHT_INFLUENCE	//only do once?
			break;
#endif
		}
	}

	//make sure to do this before adjusting weights for the next layer based upon this layer
	void ComputeHistoryTensor(void) {
		double sum;
		int i,j,j0,j2,j3;

		//this shouldn't loop back around to next_index
		//as long as we have at least 3 history entries
		int last_index = (index + n.size - 1) % n.size;

		C_RNNLayer *this_net = n[index];
		C_RNNLayer *last_net = n[last_index];

		//notice - this is not found in the tensor derivitives
		//but i figured without this, we'd be giving newer error weight (more accurate)
		//just as much reign as older error weight (which is less accurate)
		//
		//so if you dont like this idea then just set it to 1.0
		const double hist_decay = 1.0;

		//h_mem_bias

		for (j = 0; j < this_net->mem.size; j++) {		//target
			for (j2 = 0; j2 < this_net->mem.size; j2++) {	//iteration

				sum = 0.0;
				for (j3 = 0; j3 < this_net->mem.size; j3++) {
					sum += last_net->mem_mem[j3][j2] * last_net->h_mem_bias[j][j3];
				}
				sum *= hist_decay;
				if (j == j2) sum++;
				this_net->h_mem_bias[j][j2] =
					this_net->mem[j2] * (1.0 - this_net->mem[j2]) * sum;

				//h_in_mem

				for (i = 0; i < this_net->in.size; i++) {		//target in

					sum = 0.0;
					for (j3 = 0; j3 < this_net->mem.size; j3++) {
						sum += last_net->mem_mem[j3][j2] * last_net->h_in_mem[i][j][j3];
					}
					sum *= hist_decay;
					if (j == j2) sum += last_net->in[i];
					this_net->h_in_mem[i][j][j2] =
						this_net->mem[j2] * (1.0 - this_net->mem[j2]) * sum;
				}

				//h_mem_mem

				for (j0 = 0; j0 < this_net->mem.size; j0++) {	//target mem_from

					sum = 0.0;
					for (j3 = 0; j3 < this_net->mem.size; j3++) {
						sum += last_net->mem_mem[j3][j2] * last_net->h_mem_mem[j0][j][j3];
					}
					sum *= hist_decay;
					if (j == j2) sum += last_net->mem[j0];
					this_net->h_mem_mem[j0][j][j2] =
						this_net->mem[j2]
						* (1.0 - this_net->mem[j2])
						* sum;
				}
			}
		}
	}

	double Evaluate(const CVectord &in, const CVectord &des) {

		assert(index >= 0 && index < n.size);

		int i;

		//for reference. the output index of the ready signal
		int out_ready_index = n[0]->des.size-1;
		//the input index of the got input signal
		int in_gotinput_index = n[0]->in.size-1;

		//evaluate the current network (pass the next net's mem vector)
		//here's an idea:
		//designate a 'ready' signal in the output
		//eval the network up to some max times
		//if ever in the loop we get a ready signal then stop the loop
		//train the net with the last batch

		int next_index = (index + 1) % n.size;
		C_RNNLayer *this_net = n[index];
		C_RNNLayer *next_net = n[next_index];

		//minimum iterations only causes the 2-bit addr to diverge...
		//...or maybe its the >5 max iterations?
		//no, its neither.  giving us a ready signal is whats causing the network to diverge.
#if FORCE_STEPS > 0
		const int miniter = FORCE_STEPS;
		const int maxiter = FORCE_STEPS;
#else
		const int miniter = 1;
		const int maxiter = 5;
#endif
		for (int iter = 0; ; iter++) {

			//find our next network
			next_index = (index + 1) % n.size;
			next_net = n[next_index];

			//fill the input
			this_net->in.Fill(in);

#if FORCE_STEPS == 1	//hack - simulate a single-layer - zero the ->mem loopback connection
	
			this_net->mem_bias.Fill(0);
			this_net->in_mem.Fill(0);
			this_net->mem_mem.Fill(0);

#elif FORCE_STEPS == 2	//hack - simulate a double-layer - zero the in->out and the mem->mem.  use at least 2 iter

			this_net->in_out.Fill(0);
			this_net->mem_mem.Fill(0);

#endif
			//set the 'got input' signal if we're on the first iteration
			this_net->in[in_gotinput_index] = iter ? 0.1 : 0.9;

			//process the output and the next net's memory state
			this_net->Evaluate(next_net->mem);

			//this_net has all except 'des' is filled
			//next_net has only mem filled

			//now we should calculate the history tensors in this_net
			//(they're used for next_net's weight changes)
			ComputeHistoryTensor();

			//copy over the weights into the next network
			next_net->FillWeights(this_net);

			//next_net now has mem and all weights filled

			//bail if we got a ready signal
			//only do so once we've passed our minimum iteration level
			if (iter >= miniter && this_net->des[out_ready_index] >= 0.5) break;

			//if we're at the end of the loop then bail before upping the index
			if (iter >= maxiter-1) break;

			//switch to the next network
			index = next_index;
			this_net = n[index];

			//here we always loop back
		}

		//calc current network's desired vector
		////problem-specific (and a tough one at that)
		this_net->des.Fill(des);

		//calc the network error
		double error = 0;
		for (i = 0; i < this_net->out.size-1; i++) {
			double d = this_net->des[i] - this_net->out[i];
			error += 0.5 * d * d;
		}

		//set our desired ready-signal neuron to 'want a signal!'
		//
		//this is subjective and problematic.
		//will this cause our network to tend towards
		//immediately producing a ready signal?
		//
		//we want the ready signal to initially come about within 'maxiter'
		//cycles. but once that happens we want it to vary with respect to
		//the correct produced output...
		//
		//how about setting the 'desired ready signal' to a value proportional
		//to the error of the rest of the network?
		//
		//sigmoid so when the error is high, the desired tends to 1.0 to terminate quickly
		//and as the error gets lower, the desired tends to 0.5 for variation
		this_net->des[out_ready_index] = 0.1;//sigmoid(error);			//error coeff?

		//readjust for the error
		double d = this_net->des[out_ready_index] - this_net->out[out_ready_index];
		error += 0.5 * d * d;

		//adjust the weights for next_net based upon this_net
		AdjustWeights();

		//and cycle the circular buffer
		index = next_index;
		//switch to the next network
		this_net = n[index];

		//return this time's error
		return error;
	}

	void Display(bool full = false) {
		//display the last network on the list
		int this_index = (index + n.size - 1) % n.size;
		int next_index = index;
		C_RNNLayer *this_net = n[this_index];
		C_RNNLayer *next_net = n[next_index];

		if (!full) {
			this_net->Display(&next_net->mem);
		} else {
			this_net->DisplayFull(&next_net->mem);
		}
	}
};

void main(void)
{
	int iter = 0;
	double err, least, greatest;

	clrscr();
	srand(time(0));

	least = -log(0);
	greatest = -least;

	const int hist_size = 2;//10;
	const int in_size = 8;			//two-bit adder
	const int mem_size = 16;
	const int out_size = 4;

	C_RNN *rnn = new C_RNN(hist_size, in_size, mem_size, out_size);
	CVectord in(in_size);
	CVectord des(out_size);

	int pause = 0;
	do {

		if (!pause) {
			//fill in & des
			{
				int v1 = rand() ^ rand() ^ rand() ^ rand() ^ rand();
				int v2 = rand() ^ rand() ^ rand() ^ rand() ^ rand();
				int v3 = v1 + v2;
				int a;

				int bits = in.size >> 1;
				for (a = 0; a < bits; a++) {
					in[a] =			(double)((v1 & (1 << a)) == (1 << a)) * 0.8 + 0.1;
					in[a+bits] =	(double)((v2 & (1 << a)) == (1 << a)) * 0.8 + 0.1;
				}
				for (a = 0; a < des.size; a++) {
					des[a] =		(double)((v3 & (1 << a)) == (1 << a)) * 0.8 + 0.1;
				}
			}

			//todo - run 512x before displaying...
			err += rnn->Evaluate(in, des);

			printf("\n");
			rnn->Display(false);

			const int counter_max = 25;
			static int counter=0;
			static double lasterr =0;
			counter++;
			if (counter == counter_max) {
				if (err > greatest) greatest = err;
				if (err < least) least = err;
				lasterr = err;
				err = 0;
				counter = 0;
			}

			printf("smallest error: %.15f      \n", least);
			printf("current error:	%.15f      \n", lasterr);
			printf("largest error:	%.15f      \n", greatest);
			printf("\n");
			printf("total alloc: %d (0x%x)   \n", total_allocated);
		}

		char keycmd = getkeystate();
		if (!pause) {
			printf("keycmd: %d    \n", keycmd);
		}
//		if (!(keycmd & 0x80)) getch();

		if (keycmd == 1) break;
		if (keycmd == 57) {
			pause = !pause;
			getch();
			printf("getch\n");
		}

	} while (1);
}
